@@ -20,13 +20,13 @@ def check_env_tm20lidar():
     sct = mss.mss()
     monitor = {"top": 30, "left": 0, "width": 958, "height": 490}
     lidar = Lidar(monitor=monitor, road_point=(440, 479))
-    env_config = DEFAULT_CONFIG_DICT
+    env_config = DEFAULT_CONFIG_DICT.copy()
     env_config["interface"] = TM2020InterfaceLidar
     env_config["wait_on_done"] = True
     env_config["interface_kwargs"] = {"img_hist_len": 1, "gamepad": False, "min_nb_steps_before_early_done": int(20 * 60), "road_point": (440, 479), "record": False}
     # env_config["time_step_duration"] = 0.5  # nominal duration of your time-step
     # env_config["start_obs_capture"] = 0.4
-    env = gym.make("rtgym:real-time-gym-v0", config=env_config)
+    env = gym.make("real-time-gym-v0", config=env_config)
     o = env.reset()
     while True:
         o, r, d, i = env.step(None)
